@@ -1,5 +1,8 @@
 import os
 import logging
+import hashlib
+from typing import List, Optional  # pylint:disable=unused-import
+
 import sortedcontainers
 
 import archinfo
@@ -13,6 +16,68 @@ from ..errors import CLEOperationError, CLEError
 l = logging.getLogger(name=__name__)
 
 
+class FunctionHintSource:
+    """
+    Enums that describe the source of function hints.
+    """
+    EH_FRAME = 0
+    EXTERNAL_EH_FRAME = 1
+
+
+class FunctionHint:
+    """
+    Describes a function hint.
+
+    :ivar int addr:     Address of the function.
+    :ivar int size:     Size of the function.
+    :ivar source:       Source of this hint.
+    :vartype source:    int
+    """
+
+    __slots__ = ('addr', 'size', 'source')
+
+    def __init__(self, addr, size, source):
+        self.addr = addr
+        self.size = size
+        self.source = source
+
+    def __repr__(self):
+        return "<FuncHint@%#x, %d bytes>" % (self.addr, self.size)
+
+
+class ExceptionHandling:
+    """
+    Describes an exception handling.
+
+    Exception handlers are usually language-specific. In C++, it is usually implemented as try {} catch {} blocks.
+
+    :ivar int start_addr:               The beginning of the try block.
+    :ivar int size:                     Size of the try block.
+    :ivar Optional[int] handler_addr:   Address of the exception handler code.
+    :ivar type:                         Type of the exception handler. Optional.
+    :ivar Optional[int] func_addr:      Address of the function. Optional.
+    """
+
+    __slots__ = ('start_addr', 'size', 'handler_addr', 'type', 'func_addr',)
+
+    def __init__(self, start_addr, size, handler_addr=None, type_=None, func_addr=None):
+
+        self.start_addr = start_addr
+        self.size = size
+        self.handler_addr = handler_addr
+        self.type = type_
+        self.func_addr = func_addr
+
+    def __repr__(self):
+        if self.handler_addr is not None:
+            return "<ExceptionHandling@%#x-%#x: handler@%#x>" % (self.start_addr,
+                                                                 self.start_addr + self.size,
+                                                                 self.handler_addr)
+        else:
+            return "<ExceptionHandling@%#x-%#x: no handler>" % (self.start_addr,
+                                                                 self.start_addr + self.size)
+
+
 class Backend:
     """
     Main base class for CLE binary objects.
@@ -20,6 +85,7 @@ class Backend:
     An alternate interface to this constructor exists as the static method :meth:`cle.loader.Loader.load_object`
 
     :ivar binary:           The path to the file this object is loaded from
+    :ivar binary_basename:  The basename of the filepath, or a short representation of the stream it was loaded from
     :ivar is_main_bin:      Whether this binary is loaded as the main executable
     :ivar segments:         A listing of all the loaded segments in this file
     :ivar sections:         A listing of all the demarked sections in the file
@@ -49,9 +115,9 @@ class Backend:
 
     def __init__(self,
             binary,
+            binary_stream,
             loader=None,
             is_main_bin=False,
-            filename=None,
             entry_point=None,
             arch=None,
             base_addr=None,
@@ -60,17 +126,17 @@ class Backend:
             **kwargs):
         """
         :param binary:          The path to the binary to load
+        :param binary_stream:   The open stream to this binary. The reference to this will be held until you call close.
         :param is_main_bin:     Whether this binary should be loaded as the main executable
         """
-        if hasattr(binary, 'seek') and hasattr(binary, 'read'):
-            self.binary = filename
-            self.binary_stream = binary
+        self.binary = binary
+        self._binary_stream = binary_stream
+        if self.binary is not None:
+            self.binary_basename = os.path.basename(self.binary)
+        elif hasattr(self._binary_stream, "name"):
+            self.binary_basename = os.path.basename(self._binary_stream.name)
         else:
-            self.binary = binary
-            try:
-                self.binary_stream = open(binary, 'rb')
-            except IOError:
-                self.binary_stream = None
+            self.binary_basename = str(self._binary_stream)
 
         for k in list(kwargs.keys()):
             if k == 'custom_entry_point':
@@ -101,7 +167,7 @@ class Backend:
         self.jmprel = {}
         self.arch = None
         self.os = None  # Let other stuff override this
-        self.engine_preset = None
+        self.compiler = None, None  # compiler name, version
         self._symbol_cache = {}
         # a list of directories to search for libraries specified by the object
         self.extra_load_path = []
@@ -109,23 +175,37 @@ class Backend:
         self.guess_simprocs = False
         self.guess_simprocs_hint = None
 
+        # checksums
+        self.md5 = None
+        self.sha256 = None
+
         self.mapped_base_symbolic = 0
         # These are set by cle, and should not be overriden manually
         self.mapped_base = self.linked_base = 0 # not to be set manually - used by CLE
 
         self.deps = []           # Needed shared objects (libraries dependencies)
+        self.child_objects = []  # any objects loaded directly out of this
+        self.parent_object = None
         self.linking = None # Dynamic or static linking
         self.pic = force_rebase
         self.execstack = False
 
-        # tls info
+        # tls info set by backend to communicate with thread manager
         self.tls_used = False
-        self.tls_module_id = None
-        self.tls_block_offset = None
         self.tls_block_size = None
-        self.tls_data_start = None
         self.tls_data_size = None
+        self.tls_data_start = None
+        # tls info set by thread manager
+        self.tls_module_id = None
+        #self.tls_block_offset = None  # this is an ELF-only attribute
 
+        # exception handling
+        # they should be rebased when .rebase() is called
+        self.exception_handlings = []  # type: List[ExceptionHandling]
+
+        # Hints
+        # they should be rebased when .rebase() is called
+        self.function_hints = []  # type: List[FunctionHint]
 
         # Custom options
         self._custom_entry_point = entry_point
@@ -138,6 +218,10 @@ class Backend:
         self._is_mapped = False
         # cached max_addr
         self._max_addr = None
+        # cached last section
+        self._last_section = None
+        # cached last segment
+        self._last_segment = None
 
         if arch is None:
             self.arch = None
@@ -150,18 +234,14 @@ class Backend:
         else:
             raise CLEError("Bad parameter: arch=%s" % arch)
 
+        self._checksum()
+
     def close(self):
-        if self.binary_stream is not None:
-            self.binary_stream.close()
-            self.binary_stream = None
+        del self._binary_stream
 
     def __repr__(self):
-        if self.binary is not None:
-            return '<%s Object %s, maps [%#x:%#x]>' % \
-                   (self.__class__.__name__, os.path.basename(self.binary), self.min_addr, self.max_addr)
-        else:
-            return '<%s Object from stream, maps [%#x:%#x]>' % \
-                   (self.__class__.__name__, self.min_addr, self.max_addr)
+        return '<%s Object %s, maps [%#x:%#x]>' % \
+               (self.__class__.__name__, self.binary_basename, self.min_addr, self.max_addr)
 
     def set_arch(self, arch):
         self.arch = arch
@@ -208,18 +288,43 @@ class Backend:
         l.critical("Deprecation warning: symbols_by_addr is deprecated - use loader.find_symbol() for lookup and .symbols for enumeration")
         return {s.rebased_addr: s for s in self.symbols}
 
-    def rebase(self):
+    def rebase(self, new_base):
         """
         Rebase backend's regions to the new base where they were mapped by the loader
         """
         if self._is_mapped:
+            # we could rebase an object twice if we really wanted... no need though, right?
             raise CLEOperationError("Image already rebased from %#x to %#x" % (self.linked_base, self.mapped_base))
 
+        self.mapped_base = new_base
+
         if self.sections:
             self.sections._rebase(self.image_base_delta)
         if self.segments and self.sections is not self.segments:
             self.segments._rebase(self.image_base_delta)
 
+        for handling in self.exception_handlings:
+            if handling.func_addr is not None:
+                handling.func_addr += self.image_base_delta
+            if handling.handler_addr is not None:
+                handling.handler_addr += self.image_base_delta
+            handling.start_addr += self.image_base_delta
+
+        for hint in self.function_hints:
+            hint.addr = hint.addr + self.image_base_delta
+
+    def relocate(self):
+        """
+        Apply all resolved relocations to memory.
+
+        The meaning of "resolved relocations" is somewhat subtle - there is a linking step which attempts to resolve
+        each relocation, currently only present in the main internal loading function since the calculation of which
+        objects should be available
+        """
+        for reloc in self.relocs:
+            if reloc.resolved:
+                reloc.relocate()
+
     def contains_addr(self, addr):
         """
         Is `addr` in one of the binary's segments/sections we have loaded? (i.e. is it mapped into memory ?)
@@ -234,13 +339,25 @@ class Backend:
         """
         Returns the segment that contains `addr`, or ``None``.
         """
-        return self.segments.find_region_containing(addr)
+        if self._last_segment is not None and self._last_segment.contains_addr(addr):
+            return self._last_segment
+
+        r = self.segments.find_region_containing(addr)
+        if r is not None:
+            self._last_segment = r
+        return r
 
     def find_section_containing(self, addr):
         """
         Returns the section that contains `addr` or ``None``.
         """
-        return self.sections.find_region_containing(addr)
+        if self._last_section is not None and self._last_section.contains_addr(addr):
+            return self._last_section
+
+        r = self.sections.find_region_containing(addr)
+        if r is not None:
+            self._last_section = r
+        return r
 
     def addr_to_offset(self, addr):
         loadable = self.find_loadable_containing(addr)
@@ -335,6 +452,13 @@ class Backend:
         """
         return None
 
+    @classmethod
+    def is_compatible(cls, stream):  # pylint:disable=unused-argument
+        """
+        Determine quickly whether this backend can load an object from this stream
+        """
+        return False
+
     @classmethod
     def check_compatibility(cls, spec, obj): # pylint: disable=unused-argument
         """
@@ -349,17 +473,33 @@ class Backend:
         """
         return False
 
-
     @staticmethod
     def _get_symbol_relative_addr(symbol):
         return symbol.relative_addr
 
+    def _checksum(self):
+        """
+        Calculate MD5 and SHA256 checksum for the binary.
+        """
+
+        if self._binary_stream is not None:
+            data = self._binary_stream.read()
+            self._binary_stream.seek(0)
+            self.md5 = hashlib.md5(data).digest()
+            self.sha256 = hashlib.sha256(data).digest()
+
+    def __getstate__(self):
+        return self.__dict__
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        for sym in self.symbols:
+            sym.owner = self
+
 ALL_BACKENDS = dict()
 
 
 def register_backend(name, cls):
-    if not hasattr(cls, 'is_compatible'):
-        raise TypeError("Backend needs an is_compatible() method")
     ALL_BACKENDS.update({name: cls})
 
 
@@ -374,6 +514,9 @@ from .macho import MachO
 from .named_region import NamedRegion
 from .java.jar import Jar
 from .java.apk import Apk
+from .java.soot import Soot
+from .xbe import XBE
+from .static_archive import StaticArchive
 
 try:
     from .binja import BinjaBin
