@@ -1,16 +1,19 @@
+import copy
 import os
-import struct
 import logging
 import archinfo
 import elftools
 from elftools.elf import elffile, sections
+from elftools.dwarf import callframe
+from elftools.common.exceptions import ELFError, DWARFError
 from collections import OrderedDict, defaultdict
 
 from .symbol import ELFSymbol, Symbol, SymbolType
 from .regions import ELFSection, ELFSegment
 from .hashtable import ELFHashTable, GNUHashTable
 from .metaelf import MetaELF, maybedecode
-from .. import register_backend
+from .lsda import LSDAExceptionTable
+from .. import register_backend, FunctionHint, FunctionHintSource, ExceptionHandling
 from .relocation import get_relocation
 from .relocation.generic import MipsGlobalReloc, MipsLocalReloc
 from ...patched_stream import PatchedStream
@@ -29,15 +32,15 @@ class ELF(MetaELF):
     """
     is_default = True  # Tell CLE to automatically consider using the ELF backend
 
-    def __init__(self, binary, addend=None, **kwargs):
-        super(ELF, self).__init__(binary, **kwargs)
+    def __init__(self, *args, addend=None, debug_symbols=None, **kwargs):
+        super().__init__(*args, **kwargs)
         patch_undo = []
         try:
-            self.reader = elffile.ELFFile(self.binary_stream)
-            list(self.reader.iter_sections())
+            self._reader = elffile.ELFFile(self._binary_stream)
+            list(self._reader.iter_sections())
         except Exception: # pylint: disable=broad-except
-            self.binary_stream.seek(4)
-            ty = self.binary_stream.read(1)
+            self._binary_stream.seek(4)
+            ty = self._binary_stream.read(1)
             if ty not in (b'\1', b'\2'):
                 raise CLECompatibilityError
 
@@ -47,29 +50,27 @@ class ELF(MetaELF):
                 patch_data = [(0x28, b'\0'*8), (0x3a, b'\0'*6)]
 
             for offset, patch in patch_data:
-                self.binary_stream.seek(offset)
-                patch_undo.append((offset, self.binary_stream.read(len(patch))))
+                self._binary_stream.seek(offset)
+                patch_undo.append((offset, self._binary_stream.read(len(patch))))
 
-            self.binary_stream = PatchedStream(self.binary_stream, patch_data)
+            self._binary_stream = PatchedStream(self._binary_stream, patch_data)
             l.error("PyReadELF couldn't load this file. Trying again without section headers...")
 
             try:
-                self.reader = elffile.ELFFile(self.binary_stream)
+                self._reader = elffile.ELFFile(self._binary_stream)
             except Exception as e: # pylint: disable=broad-except
                 raise CLECompatibilityError from e
 
         # Get an appropriate archinfo.Arch for this binary, unless the user specified one
         if self.arch is None:
-            self.set_arch(self.extract_arch(self.reader))
+            self.set_arch(self.extract_arch(self._reader))
         else:
-            other_arch = self.extract_arch(self.reader)
+            other_arch = self.extract_arch(self._reader)
             if other_arch != self.arch:
                 l.warning("User specified %s but autodetected %s. Proceed with caution.", self.arch, other_arch)
 
         self._addend = addend
 
-        self.strtab = None
-        self.dynsym = None
         self.hashtable = None
 
         self._dynamic = {}
@@ -90,13 +91,20 @@ class ELF(MetaELF):
         self._desperate_for_symbols = False
         self.imports = {}
         self.resolved_imports = []
+        self.tls_block_offset = None  # this is an ELF-only attribute
 
         self.relocs = []
         self.jmprel = OrderedDict()
 
-        self._entry = self.reader.header.e_entry
-        self.is_relocatable = self.reader.header.e_type == 'ET_REL'
-        self.pic = self.pic or self.reader.header.e_type in ('ET_REL', 'ET_DYN')
+        #
+        # DWARF data
+        #
+        self.has_dwarf_info = bool(self._reader.has_dwarf_info())
+        self.build_id = None
+
+        self._entry = self._reader.header.e_entry
+        self.is_relocatable = self._reader.header.e_type == 'ET_REL'
+        self.pic = self.pic or self._reader.header.e_type in ('ET_REL', 'ET_DYN')
 
         self.__parsed_reloc_tables = set()
 
@@ -104,12 +112,12 @@ class ELF(MetaELF):
         # the fact that elffile, used by those methods, is working only with un-based virtual addresses, but Clemories
         # themselves are organized as a tree where each node backer internally uses relative addressing
         seg_addrs = (ALIGN_DOWN(x['p_vaddr'], self.loader.page_size)
-                     for x in self.reader.iter_segments() if x.header.p_type == 'PT_LOAD' and x.header.p_memsz > 0)
+                     for x in self._reader.iter_segments() if x.header.p_type == 'PT_LOAD' and x.header.p_memsz > 0)
         self.mapped_base = self.linked_base = 0
         try:
             self.mapped_base = self.linked_base = min(seg_addrs)
         except ValueError:
-            l.warning('no segments identified in PT_LOAD')
+            l.info('no PT_LOAD segments identified')
 
         self.__register_segments()
         self.__register_sections()
@@ -118,6 +126,27 @@ class ELF(MetaELF):
             self._desperate_for_symbols = True
             self.symbols.update(self._symbol_cache.values())
 
+        if self.has_dwarf_info and self.loader._load_debug_info:
+            # load DWARF information
+            try:
+                dwarf = self._reader.get_dwarf_info()
+            except ELFError:
+                l.warning("An exception occurred in pyelftools when loading the DWARF information for %s. "
+                          "Marking DWARF as not available for this binary.", self.binary_basename, exc_info=True)
+                dwarf = None
+                self.has_dwarf_info = False
+
+            if dwarf and dwarf.has_EH_CFI():
+                self._load_function_hints_from_fde(dwarf, FunctionHintSource.EH_FRAME)
+                self._load_exception_handling(dwarf)
+
+        if debug_symbols:
+            self.__process_debug_file(debug_symbols)
+        elif self.loader._load_debug_info and self.build_id:
+            debug_filename = '/usr/lib/debug/.build-id/%s/%s.debug' % (self.build_id[:2], self.build_id[2:])
+            if os.path.isfile(debug_filename):
+                self.__process_debug_file(debug_filename)
+
         # call the methods defined by MetaELF
         self._ppc64_abiv1_entry_fix()
         self._load_plt()
@@ -126,11 +155,14 @@ class ELF(MetaELF):
             self.memory.store(AT.from_lva(self.min_addr + offset, self).to_rva(), patch)
 
 
-
     #
     # Properties and Public Methods
     #
 
+    def close(self):
+        super().close()
+        del self._reader
+
     @classmethod
     def check_compatibility(cls, spec, obj):
         with stream_or_path(spec) as stream:
@@ -239,17 +271,16 @@ class ELF(MetaELF):
 
         :param symid: Either an index into .dynsym or the name of a symbol.
         """
-        if symbol_table is None:
-            symbol_table = self.dynsym
-
         if type(symid) is int:
             if symid == 0:
                 # special case the null symbol, this is important for static binaries
                 return self._nullsymbol
+            if symbol_table is None:
+                raise TypeError("Must specify the symbol table to look up symbols by index")
             try:
                 re_sym = symbol_table.get_symbol(symid)
             except Exception:  # pylint: disable=broad-except
-                l.exception("Error parsing symbol at %#08x", symid)
+                l.exception("Error parsing symbol %#08x", symid)
                 return None
             cache_key = self._symbol_to_tuple(re_sym)
             cached = self._symbol_cache.get(cache_key, None)
@@ -339,18 +370,9 @@ class ELF(MetaELF):
         self._inits_extracted = True
 
     def _load_segment(self, seg):
-        self._load_segment_metadata(seg)
-        self._load_segment_memory(seg)
-
-    def _load_segment_metadata(self, seg):
-        """
-        Loads a segment based on a LOAD directive in the program header table.
-        """
         loaded_segment = ELFSegment(seg)
         self.segments.append(loaded_segment)
 
-    def _load_segment_memory(self, seg):
-
         # see https://code.woboq.org/userspace/glibc/elf/dl-load.c.html#1066
         ph = seg.header
 
@@ -370,11 +392,14 @@ class ELF(MetaELF):
 
         mapoff = ALIGN_DOWN(ph.p_offset, self.loader.page_size)
 
+        # patch modified addresses into ELFSegment
+        loaded_segment.vaddr = mapstart
+        loaded_segment.memsize = mapend - mapstart
+        loaded_segment.filesize = mapend - mapstart
+        loaded_segment.offset = mapoff
+
         # see https://code.woboq.org/userspace/glibc/elf/dl-map-segments.h.html#88
         data = get_mmaped_data(seg.stream, mapoff, mapend - mapstart, self.loader.page_size)
-        if not data:
-            l.warning("Segment %s is empty at %#08x!", seg.header.p_type, mapstart)
-            return
         if allocend > dataend:
             zero = dataend
             zeropage = (zero + self.loader.page_size - 1) & ~(self.loader.page_size - 1)
@@ -385,6 +410,10 @@ class ELF(MetaELF):
             zeroend = ALIGN_UP(allocend, self.loader.page_size) # mmap maps to the next page boundary
             if zeroend > zeropage:
                 data = data.ljust(zeroend - mapstart, b'\0')
+            loaded_segment.memsize = zeroend - mapstart
+        elif not data:
+            l.warning("Segment %s is empty at %#08x!", seg.header.p_type, mapstart)
+            return
 
         self.memory.add_backer(AT.from_lva(mapstart, self).to_rva(), data)
 
@@ -398,75 +427,84 @@ class ELF(MetaELF):
         if dest_section is not None:
             address += dest_section.remap_offset
 
-        return RelocClass(self, symbol, address, addend)
+        try:
+            return RelocClass(self, symbol, address, addend)
+        except KeyError:
+            l.error("Malformed relocation: access to unmapped %#x", readelf_reloc.entry.r_offset)
+            return None
+
+    def _load_function_hints_from_fde(self, dwarf, source):
+        """
+        Load frame description entries out of the .eh_frame section. These entries include function addresses and can be
+        used to improve CFG recovery.
+
+        :param dwarf:   The DWARF info object from pyelftools.
+        :return:        None
+        """
+
+        try:
+            for entry in dwarf.EH_CFI_entries():
+                if type(entry) is callframe.FDE:
+                    self.function_hints.append(FunctionHint(
+                        entry.header['initial_location'],
+                        entry.header['address_range'],
+                        source,
+                    ))
+        except (DWARFError, ValueError):
+            l.warning("An exception occurred in pyelftools when loading FDE information.",
+                      exc_info=True)
+
+    def _load_exception_handling(self, dwarf):
+        """
+        Load exception handling information out of the .eh_frame and .gcc_except_table sections. We may support more
+        types of exception handling information in the future.
+
+        :param dwarf:   The DWARF info object from pyelftools.
+        :return:        None
+        """
+
+        try:
+            lsda = LSDAExceptionTable(self._binary_stream, self.arch.bits, self.arch.memory_endness == 'Iend_LE')
+            for entry in dwarf.EH_CFI_entries():
+                if type(entry) is callframe.FDE and \
+                        hasattr(entry, 'lsda_pointer') and \
+                        entry.lsda_pointer is not None:
+                    # function address
+                    func_addr = entry.header.get('initial_location', None)
+                    if func_addr is None:
+                        l.warning('Unexpected FDE structure: '
+                                  'initial_location is not specified while LSDA pointer is available.')
+                        continue
+
+                    # Load and parse LSDA exception table
+                    file_offset = self.addr_to_offset(entry.lsda_pointer)
+                    lsda_exception_table = lsda.parse_lsda(entry.lsda_pointer, file_offset)
+                    for exc in lsda_exception_table:
+                        handling = ExceptionHandling(
+                            func_addr + exc.cs_start,
+                            exc.cs_len,
+                            handler_addr=func_addr + exc.cs_lp if exc.cs_lp != 0 else None,
+                            func_addr=func_addr
+                        )
+                        self.exception_handlings.append(handling)
+
+        except (DWARFError, ValueError):
+            l.warning("An exception occurred in pyelftools when loading FDE information.",
+                      exc_info=True)
 
     #
     # Private Methods... really. Calling these out of context
     # will probably break things. Caveat emptor.
     #
 
-    def __getstate__(self):
-        if self.binary is None:
-            raise ValueError("Can't pickle an object loaded from a stream")
-
-        state = dict(self.__dict__)
-
-        # Trash the unpickleable
-        if type(self.binary_stream) is PatchedStream:
-            state['binary_stream'].stream = None
-        else:
-            state['binary_stream'] = None
-
-        state['reader'] = None
-        state['strtab'] = None
-        state['dynsym'] = None
-        state['hashtable'] = None
-
-        return state
-
-    def __setstate__(self, data):
-        self.__dict__.update(data)
-
-        if self.binary_stream is None:
-            self.binary_stream = open(self.binary, 'rb')
-        else:
-            self.binary_stream.stream = open(self.binary, 'rb')
-
-        self.reader = elffile.ELFFile(self.binary_stream)
-        if self._dynamic and 'DT_STRTAB' in self._dynamic:
-            self.strtab = next(x for x in self.reader.iter_segments() if x.header.p_type == 'PT_DYNAMIC')._get_stringtable()
-            if 'DT_SYMTAB' in self._dynamic and 'DT_SYMENT' in self._dynamic:
-                fakesymtabheader = {
-                    'sh_offset': AT.from_lva(self._dynamic['DT_SYMTAB'], self).to_rva(),
-                    'sh_entsize': self._dynamic['DT_SYMENT'],
-                    'sh_size': 0, # bogus size: no iteration allowed
-                    'sh_flags': 0,
-                    'sh_addralign': 0,
-                }
-                self.dynsym = elffile.SymbolTableSection(fakesymtabheader, 'symtab_cle', self.reader, self.strtab)
-                self.dynsym.stream = self.memory
-
-                if 'DT_GNU_HASH' in self._dynamic:
-                    self.hashtable = GNUHashTable(
-                            self.dynsym,
-                            self.memory,
-                            AT.from_lva(self._dynamic['DT_GNU_HASH'], self).to_rva(),
-                            self.arch)
-                elif 'DT_HASH' in self._dynamic:
-                    self.hashtable = ELFHashTable(
-                            self.dynsym,
-                            self.memory,
-                            AT.from_lva(self._dynamic['DT_HASH'], self).to_rva(),
-                            self.arch)
-
     def __register_segments(self):
         self.linking = 'static'
-        if self.is_relocatable and self.reader.num_segments() > 0:
+        if self.is_relocatable and self._reader.num_segments() > 0:
             # WTF?
             raise CLEError("Relocatable objects with segments are not supported.")
 
         type_to_seg_mapping = defaultdict(list)
-        for seg_readelf in self.reader.iter_segments():
+        for seg_readelf in self._reader.iter_segments():
             type_to_seg_mapping[seg_readelf.header.p_type].append(seg_readelf)
 
         # PT_LOAD segments must be processed first so that the memory_backers for the other segments exist
@@ -484,6 +522,8 @@ class ELF(MetaELF):
         for seg in type_to_seg_mapping['PT_GNU_STACK']:
             self.execstack = bool(seg.header.p_flags & 1)
 
+        for seg in type_to_seg_mapping['PT_GNU_RELRO']:
+            self.__register_relro(seg)
 
     def __register_dyn(self, seg_readelf):
         """
@@ -506,9 +546,11 @@ class ELF(MetaELF):
 
         self.extra_load_path = self.__parse_rpath(runpath, rpath)
 
-        self.strtab = seg_readelf._get_stringtable()
+        strtab = seg_readelf._get_stringtable()
+        self.__neuter_streams(strtab)
 
         # To extract symbols from binaries without section headers, we need to hack into pyelftools.
+        # TODO: pyelftools is less bad than it used to be. how much of this can go away?
         # None of the following things make sense without a string table or a symbol table
         if 'DT_STRTAB' in self._dynamic and 'DT_SYMTAB' in self._dynamic and 'DT_SYMENT' in self._dynamic:
                 # Construct our own symbol table to hack around pyreadelf assuming section headers are around
@@ -519,41 +561,42 @@ class ELF(MetaELF):
                     'sh_flags': 0,
                     'sh_addralign': 0,
                 }
-                self.dynsym = elffile.SymbolTableSection(fakesymtabheader, 'symtab_cle', self.reader, self.strtab)
-                self.dynsym.stream = self.memory
+                dynsym = elffile.SymbolTableSection(fakesymtabheader, 'symtab_cle', self._reader, strtab)
+                dynsym.stream = self.memory
+                dynsym.elffile = None
 
-                # set up the hash table, prefering the gnu hash section to the old hash section
+                # set up the hash table, preferring the gnu hash section to the old hash section
                 # the hash table lets you get any symbol given its name
                 if 'DT_GNU_HASH' in self._dynamic:
                     self.hashtable = GNUHashTable(
-                        self.dynsym, self.memory, AT.from_lva(self._dynamic['DT_GNU_HASH'], self).to_rva(), self.arch)
+                        dynsym, self.memory, AT.from_lva(self._dynamic['DT_GNU_HASH'], self).to_rva(), self.arch)
                 elif 'DT_HASH' in self._dynamic:
                     self.hashtable = ELFHashTable(
-                        self.dynsym, self.memory, AT.from_lva(self._dynamic['DT_HASH'], self).to_rva(), self.arch)
+                        dynsym, self.memory, AT.from_lva(self._dynamic['DT_HASH'], self).to_rva(), self.arch)
                 else:
                     l.warning("No hash table available in %s", self.binary)
 
                 # mips' relocations are absolutely screwed up, handle some of them here.
-                self.__relocate_mips()
+                self.__relocate_mips(dynsym)
 
                 # perform a lot of checks to figure out what kind of relocation tables are around
                 self.rela_type = None
                 if 'DT_PLTREL' in self._dynamic:
                     if self._dynamic['DT_PLTREL'] == 7:
                         self.rela_type = 'RELA'
-                        relentsz = self.reader.structs.Elf_Rela.sizeof()
+                        relentsz = self._reader.structs.Elf_Rela.sizeof()
                     elif self._dynamic['DT_PLTREL'] == 17:
                         self.rela_type = 'REL'
-                        relentsz = self.reader.structs.Elf_Rel.sizeof()
+                        relentsz = self._reader.structs.Elf_Rel.sizeof()
                     else:
                         raise CLEInvalidBinaryError('DT_PLTREL is not REL or RELA?')
                 else:
                     if 'DT_RELA' in self._dynamic:
                         self.rela_type = 'RELA'
-                        relentsz = self.reader.structs.Elf_Rela.sizeof()
+                        relentsz = self._reader.structs.Elf_Rela.sizeof()
                     elif 'DT_REL' in self._dynamic:
                         self.rela_type = 'REL'
-                        relentsz = self.reader.structs.Elf_Rel.sizeof()
+                        relentsz = self._reader.structs.Elf_Rel.sizeof()
                     else:
                         return
 
@@ -573,9 +616,10 @@ class ELF(MetaELF):
                         'sh_flags': 0,
                         'sh_addralign': 0,
                     }
-                    readelf_relocsec = elffile.RelocationSection(fakerelheader, 'reloc_cle', self.reader)
+                    readelf_relocsec = elffile.RelocationSection(fakerelheader, 'reloc_cle', self._reader)
                     readelf_relocsec.stream = self.memory
-                    self.__register_relocs(readelf_relocsec)
+                    readelf_relocsec.elffile = None
+                    self.__register_relocs(readelf_relocsec, dynsym)
 
                 # try to parse relocations out of a table of type DT_JMPREL
                 if 'DT_JMPREL' in self._dynamic:
@@ -591,9 +635,10 @@ class ELF(MetaELF):
                         'sh_flags': 0,
                         'sh_addralign': 0,
                     }
-                    readelf_jmprelsec = elffile.RelocationSection(fakejmprelheader, 'jmprel_cle', self.reader)
+                    readelf_jmprelsec = elffile.RelocationSection(fakejmprelheader, 'jmprel_cle', self._reader)
                     readelf_jmprelsec.stream = self.memory
-                    self.__register_relocs(readelf_jmprelsec, force_jmprel=True)
+                    readelf_jmprelsec.elffile = None
+                    self.__register_relocs(readelf_jmprelsec, dynsym, force_jmprel=True)
 
     def __parse_rpath(self, runpath, rpath):
         """
@@ -601,7 +646,7 @@ class ELF(MetaELF):
         """
         # DT_RUNPATH takes predence over DT_RPATH
         if len(runpath) > 0:
-            runpath = runpath
+            pass
         else:
             runpath = rpath
 
@@ -615,12 +660,14 @@ class ELF(MetaELF):
             parts.append(part)
         return parts
 
-    def __register_relocs(self, section, force_jmprel=False):
+    def __register_relocs(self, section, dynsym=None, force_jmprel=False):
+        if not self.loader._perform_relocations:
+            return
 
         got_min = got_max = 0
 
         if not force_jmprel:
-            got_sec = self.reader.get_section_by_name('.got')
+            got_sec = self._reader.get_section_by_name('.got')
             if got_sec is not None:
                 got_min = got_sec.header.sh_addr
                 got_max = got_min + got_sec.header.sh_size
@@ -642,27 +689,21 @@ class ELF(MetaELF):
                     # The target section is not loaded into memory, so just continue
                     return
 
-        symtab = self.reader.get_section(section.header['sh_link']) if 'sh_link' in section.header else None
+        symtab = self._reader.get_section(section.header['sh_link']) if 'sh_link' in section.header else dynsym
         if isinstance(symtab, elftools.elf.sections.NullSection):
             # Oh my god Atmel please stop
-            symtab = self.reader.get_section_by_name('.symtab')
+            symtab = self._reader.get_section_by_name('.symtab')
         relocs = []
         for readelf_reloc in section.iter_relocations():
             # MIPS64 is just plain old fucked up
             # https://www.sourceware.org/ml/libc-alpha/2003-03/msg00153.html
             if self.arch.name == 'MIPS64':
-                # Little endian additionally needs one of its fields reversed... WHY
-                if self.arch.memory_endness == 'Iend_LE':
-                    readelf_reloc.entry.r_info_sym = readelf_reloc.entry.r_info & 0xFFFFFFFF
-                    readelf_reloc.entry.r_info = struct.unpack('>Q', struct.pack('<Q', readelf_reloc.entry.r_info))[0]
-
-                type_1 = readelf_reloc.entry.r_info & 0xFF
-                type_2 = readelf_reloc.entry.r_info >> 8 & 0xFF
-                type_3 = readelf_reloc.entry.r_info >> 16 & 0xFF
-                extra_sym = readelf_reloc.entry.r_info >> 24 & 0xFF
-                if extra_sym != 0:
-                    l.error('r_info_extra_sym is nonzero??? PLEASE SEND HELP')
+                if not hasattr(readelf_reloc.entry, 'r_info_type2') and hasattr(readelf_reloc.entry, 'r_info_type3'):
+                    raise CLECompatibilityError("This code relies on `pyelftools` features that are not available on versions 0.26 and downwards.")
 
+                type_1 = readelf_reloc.entry.r_info_type
+                type_2 = readelf_reloc.entry.r_info_type2
+                type_3 = readelf_reloc.entry.r_info_type3
                 symbol = self.get_symbol(readelf_reloc.entry.r_info_sym, symtab)
 
                 if type_1 != 0:
@@ -702,11 +743,70 @@ class ELF(MetaELF):
         self.tls_data_size = seg_readelf.header.p_filesz
         self.tls_data_start = AT.from_lva(seg_readelf.header.p_vaddr, self).to_rva()
 
+    def __register_relro(self, segment_relro):
+        segment_relro = ELFSegment(segment_relro, relro=True)
+        vaddr = ALIGN_DOWN(segment_relro.vaddr, self.loader.page_size)
+        vaddr_end = ALIGN_UP(vaddr + segment_relro.memsize, self.loader.page_size)
+        vaddr_endfile = ALIGN_UP(vaddr + segment_relro.filesize, self.loader.page_size)
+
+        segment_relro.offset = ALIGN_DOWN(segment_relro.offset, self.loader.page_size)
+        segment_relro.vaddr = vaddr
+        segment_relro.memsize = vaddr_end - vaddr
+        segment_relro.filesize = vaddr_endfile - vaddr
+
+        def ___segments_overlap(seg1, seg2):
+            # Re-arrange so seg1 starts first
+            seg1, seg2 = (seg1, seg2) if seg1.min_addr < seg2.min_addr else (seg2, seg1)
+            # seg1 and seg2 overlap if seg2 starts before seg1 ends
+            return seg2.min_addr <= seg1.max_addr
+
+        overlapping_segments = [seg for seg in self.segments if ___segments_overlap(segment_relro, seg)]
+
+        if len(overlapping_segments) == 0:
+            l.error("RELRO segment does not overlap with any loaded segment.")
+            return
+
+        if len(overlapping_segments) > 1:
+            # I don't think this ever happens.  If it does, weshould
+            # probably also split the RELRO segment so that each one
+            # has the right permissions in case the two overlapping
+            # segments have different permissions.
+            l.warning("RELRO segment overlaps multiple segments.")
+
+        for overlapping_segment in overlapping_segments:
+            # We will split the overlapping segment into two pieces:
+            # one for the segment below the RELRO segment, and one for
+            # above.
+            segment_below = copy.copy(overlapping_segment)
+            segment_below.memsize = segment_relro.min_addr - overlapping_segment.min_addr
+            segment_below.filesize = max(segment_below.filesize, segment_below.memsize)
+
+            segment_above = copy.copy(overlapping_segment)
+            segment_above.vaddr = segment_relro.max_addr + 1
+            segment_above.memsize = overlapping_segment.max_addr - segment_above.vaddr + 1
+            segment_above.filesize = max(0, segment_above.filesize - segment_relro.memsize - segment_below.memsize)
+            segment_above.offset = segment_above.offset + segment_below.memsize + segment_relro.memsize
+
+            split_segments = [segment_below, segment_relro, segment_above]
+            split_segments = [seg for seg in split_segments if seg.memsize > 0]
+
+            # Remove the original segment
+            self.segments.remove(overlapping_segment)
+
+            # Add the new ones
+            for seg in split_segments:
+                self.segments.append(seg)
+
+        # Add the actual relro segment, and mark it as always
+        # read-only.  We use the flags of the overlapping original
+        # segment.
+        segment_relro.flags = overlapping_segments[0].flags & ~2
+
     def __register_sections(self):
         new_addr = 0
         sec_list = []
 
-        for sec_readelf in self.reader.iter_sections():
+        for sec_readelf in self._reader.iter_sections():
             remap_offset = 0
             if self.is_relocatable and sec_readelf.header['sh_flags'] & 2:      # alloc flag
                 # Relocatable objects' section addresses are meaningless (they are meant to be relocated anyway)
@@ -731,7 +831,7 @@ class ELF(MetaELF):
                 self.__register_section_symbols(sec_readelf)
             if isinstance(sec_readelf, elffile.RelocationSection) and not \
                     ('DT_REL' in self._dynamic or 'DT_RELA' in self._dynamic or 'DT_JMPREL' in self._dynamic):
-                self.__register_relocs(sec_readelf)
+                self.__register_relocs(sec_readelf, dynsym=None)
 
             if section.occupies_memory:      # alloc flag - stick in memory maybe!
                 if AT.from_lva(section.vaddr, self).to_rva() not in self.memory:        # only allocate if not already allocated (i.e. by program header)
@@ -740,11 +840,46 @@ class ELF(MetaELF):
                     else: #elif section.type == 'SHT_PROGBITS':
                         self.memory.add_backer(AT.from_lva(section.vaddr, self).to_rva(), sec_readelf.data())
 
+            if sec_readelf.header.sh_type == 'SHT_NOTE':
+                self.__register_notes(sec_readelf)
+
+            if section.name == '.comment':
+                self.__analyze_comments(sec_readelf.data())
+
+    def __register_notes(self, sec_readelf):
+        for note in sec_readelf.iter_notes():
+            if note.n_type == 'NT_GNU_BUILD_ID' and note.n_name == 'GNU':
+                if self.build_id is not None and self.build_id != note.n_desc:
+                    l.error("Mismatched build IDs present")
+                self.build_id = note.n_desc
+
+    def __analyze_comments(self, data):
+        try:
+            data = data.decode().split('\0')
+        except UnicodeDecodeError:
+            return
+
+        for line in data:
+            versions = [word for word in line.replace('(', ' ').replace(')', ' ').split() if '.' in word]
+            if not versions:
+                continue
+            versions.sort(key=len)
+            version = versions[-1]
+            lline = line.lower()
+            if 'clang' in lline:
+                compiler = 'clang'
+            elif 'gcc' in lline:
+                compiler = 'gcc'
+            else:
+                continue
+
+            self.compiler = compiler, version
+
     def __register_section_symbols(self, sec_re):
         for sym_re in sec_re.iter_symbols():
             self.symbols.add(self.get_symbol(sym_re))
 
-    def __relocate_mips(self):
+    def __relocate_mips(self, symtab):
         if 'DT_MIPS_BASE_ADDRESS' not in self._dynamic:
             return False
         # The MIPS GOT is an array of addresses, simple as that.
@@ -759,10 +894,57 @@ class ELF(MetaELF):
             self.relocs.append(reloc)
 
         for i in range(symbol_count - symtab_got_idx):
-            symbol = self.get_symbol(i + symtab_got_idx)
+            symbol = self.get_symbol(i + symtab_got_idx, symtab)
             reloc = MipsGlobalReloc(self, symbol, gotaddr + (i + got_local_num)*wordsize)
             self.relocs.append(reloc)
             self.jmprel[symbol.name] = reloc
         return True
 
+    def __neuter_streams(self, obj):
+        if isinstance(obj, elftools.elf.dynamic._DynamicStringTable):
+           obj._stream = self.memory
+           obj._table_offset = self._offset_to_rva(obj._table_offset)
+        elif isinstance(obj, elftools.elf.sections.Section):
+            if obj.header.sh_type == 'SHT_NOBITS':
+                obj.stream = None
+                obj.elffile = None
+                obj.header.sh_offset = None
+            else:
+                obj.stream = self.memory
+                obj.elffile = None
+                obj.header.sh_offset = self._offset_to_rva(obj.header.sh_offset)
+        else:
+            raise TypeError("Can't convert %r" % type(obj))
+
+    def _offset_to_rva(self, offset):
+        return AT.from_mva(self.offset_to_addr(offset), self).to_rva()
+
+    def __process_debug_file(self, filename):
+        with open(filename, 'rb') as fp:
+            try:
+                elf = elffile.ELFFile(fp)
+            except ELFError:
+                l.warning("pyelftools failed to load debug file %s", filename, exc_info=True)
+                return
+
+            for sec_readelf in elf.iter_sections():
+                if isinstance(sec_readelf, elffile.SymbolTableSection):
+                    self.__register_section_symbols(sec_readelf)
+                elif sec_readelf.header.sh_type == 'SHT_NOTE':
+                    self.__register_notes(sec_readelf)
+
+            #has_dwarf_info = bool(elf.has_dwarf_info())
+            #if has_dwarf_info:
+            #    try:
+            #        dwarf = self.reader.get_dwarf_info()
+            #    except ELFError:
+            #        l.warning("An exception occurred in pyelftools when loading the DWARF information on %s.", filename,
+            #                  exc_info=True)
+            #        dwarf = None
+
+                # debug symbols don't have eh_frame ever from what I can tell
+                #if dwarf and dwarf.has_EH_CFI():
+                #    self._load_function_hints_from_fde(dwarf, FunctionHintSource.EXTERNAL_EH_FRAME)
+
+
 register_backend('elf', ELF)
