@@ -1,8 +1,7 @@
-
 import logging
 
 from ....address_translator import AT
-from ....errors import CLEOperationError
+from ....errors import CLEOperationError, CLEInvalidBinaryError
 from ... import SymbolType
 from .elfreloc import ELFReloc
 
@@ -14,52 +13,50 @@ class GenericTLSDoffsetReloc(ELFReloc):
     def value(self):
         return self.addend + self.symbol.relative_addr
 
-    def resolve_symbol(self, solist, bypass_compatibility=False, thumb=False):  #pylint: disable=unused-argument
+    def resolve_symbol(self, solist, **kwargs):  #pylint: disable=unused-argument
         self.resolve(None)
         return True
 
 
 class GenericTLSOffsetReloc(ELFReloc):
-    def relocate(self, solist, bypass_compatibility=False):  #pylint: disable=unused-argument
+    AUTO_HANDLE_NONE = True
+
+    def relocate(self):
         hell_offset = self.owner.arch.elf_tls.tp_offset
-        if self.symbol.type == SymbolType.TYPE_NONE:
-            self.owner.memory.pack_word(
-                self.relative_addr,
-                self.owner.tls_block_offset + self.addend + self.symbol.relative_addr - hell_offset)
-            self.resolve(None)
+
+        if self.resolvedby is None:
+            obj = self.owner
         else:
-            if not self.resolve_symbol(solist, bypass_compatibility):
-                return False
-            self.owner.memory.pack_word(
-                self.relative_addr,
-                self.resolvedby.owner.tls_block_offset + self.addend + self.symbol.relative_addr - hell_offset)
-        return True
+            obj = self.resolvedby.owner
+
+        if obj.tls_block_offset is None:
+            raise CLEInvalidBinaryError("Illegal relocation - dynamically loaded object using static TLS")
+
+        self.owner.memory.pack_word(
+            self.relative_addr,
+            obj.tls_block_offset + self.addend + self.symbol.relative_addr - hell_offset)
 
 
 class GenericTLSModIdReloc(ELFReloc):
-    def relocate(self, solist, bypass_compatibility=False):  #pylint: disable=unused-argument
+    AUTO_HANDLE_NONE = True
+
+    def relocate(self):
         if self.symbol.type == SymbolType.TYPE_NONE:
-            self.owner.memory.pack_word(self.relative_addr, self.owner.tls_module_id)
-            self.resolve(None)
+            obj = self.owner
         else:
-            if not self.resolve_symbol(solist):
-                return False
-            self.owner.memory.pack_word(self.relative_addr, self.resolvedby.owner.tls_module_id)
-        return True
+            obj = self.resolvedby.owner
+
+        self.owner.memory.pack_word(self.relative_addr, obj.tls_module_id)
 
 
 class GenericIRelativeReloc(ELFReloc):
-    def relocate(self, solist, bypass_compatibility=False):  #pylint: disable=unused-argument
+    AUTO_HANDLE_NONE = True
+
+    def relocate(self):
         if self.symbol.type == SymbolType.TYPE_NONE:
             self.owner.irelatives.append((AT.from_lva(self.addend, self.owner).to_mva(), self.relative_addr))
-            self.resolve(None)
-            return True
-
-        if not self.resolve_symbol(solist, bypass_compatibility):
-            return False
-
-        self.owner.irelatives.append((self.resolvedby.rebased_addr, self.relative_addr))
-        return True
+        else:
+            self.owner.irelatives.append((self.resolvedby.rebased_addr, self.relative_addr))
 
 
 class GenericAbsoluteAddendReloc(ELFReloc):
@@ -84,22 +81,14 @@ class GenericJumpslotReloc(ELFReloc):
 
 
 class GenericRelativeReloc(ELFReloc):
+    AUTO_HANDLE_NONE = True
+
     @property
     def value(self):
         if self.resolvedby is not None:
             return self.resolvedby.rebased_addr
         return self.owner.mapped_base + self.addend
 
-    def resolve_symbol(self, solist, bypass_compatibility=False, thumb=False):  #pylint: disable=unused-argument
-        if self.symbol.type == SymbolType.TYPE_NONE:
-            self.resolve(None)
-            return True
-        return super(GenericRelativeReloc, self).resolve_symbol(
-            solist,
-            bypass_compatibility=bypass_compatibility,
-            thumb=thumb
-        )
-
 
 class GenericAbsoluteReloc(ELFReloc):
     @property
@@ -108,15 +97,11 @@ class GenericAbsoluteReloc(ELFReloc):
 
 
 class GenericCopyReloc(ELFReloc):
-    def relocate(self, solist, bypass_compatibility=False):
-        try:
-            solist.remove(self.owner)
-        except ValueError:
-            pass
-
-        if not self.resolve_symbol(solist, bypass_compatibility):
-            return False
+    def resolve_symbol(self, solist, **kwargs):
+        new_solist = [x for x in solist if x is not self.owner]
+        super().resolve_symbol(new_solist, **kwargs)
 
+    def relocate(self):
         if self.resolvedby.size != self.symbol.size and (self.resolvedby.size != 0 or not self.resolvedby.is_extern):
             l.error("Export symbol is different size than import symbol for copy relocation: %s", self.symbol.name)
         else:
@@ -129,19 +114,22 @@ class MipsGlobalReloc(GenericAbsoluteReloc):
 
 
 class MipsLocalReloc(ELFReloc):
-    def relocate(self, solist, bypass_compatibility=False): # pylint: disable=unused-argument
+    AUTO_HANDLE_NONE = True
+
+    def resolve_symbol(self, solist, **kwargs):
+        self.resolve(None)
+
+    def relocate(self):
         if self.owner.mapped_base == 0:
-            self.resolve(None)
-            return True                     # don't touch local relocations on the main bin
+            return   # don't touch local relocations on the main bin
+
         delta = self.owner.mapped_base - self.owner._dynamic['DT_MIPS_BASE_ADDRESS']
         if delta == 0:
-            self.resolve(None)
-            return True
+            return
+
         val = self.owner.memory.unpack_word(self.relative_addr)
         newval = val + delta
         self.owner.memory.pack_word(self.relative_addr, newval)
-        self.resolve(None)
-        return True
 
 
 class RelocTruncate32Mixin:
@@ -155,10 +143,7 @@ class RelocTruncate32Mixin:
     # If True, 32-bit truncated value must equal to its original when sign-extended
     check_sign_extend = False
 
-    def relocate(self, solist, bypass_compatibility=False): # pylint: disable=unused-argument
-        if not self.resolve_symbol(solist):
-            return False
-
+    def relocate(self):
         arch_bits = self.owner.arch.bits
         assert arch_bits >= 32            # 16-bit makes no sense here
 
@@ -173,3 +158,15 @@ class RelocTruncate32Mixin:
         self.owner.memory.pack_word(self.dest_addr, val, size=4, signed=False)
 
         return True
+
+
+class RelocGOTMixin:
+    """
+    A mix-in class which will cause the symbol to be resolved to a pointer to the symbol instead of the symbol
+    """
+
+    def resolve(self, symbol, extern_object=None, **kwargs):
+        assert extern_object is not None, "I have no idea how this would happen"
+
+        got_symbol = extern_object.make_extern('got.%s' % symbol.name, sym_type=SymbolType.TYPE_OBJECT, point_to=symbol)
+        super().resolve(got_symbol)
