@@ -3,9 +3,8 @@ import logging
 import binascii
 import struct
 
-from . import register_backend
+from . import register_backend, Backend
 from ..errors import CLEError
-from .blob import Blob
 
 l = logging.getLogger(name=__name__)
 
@@ -27,7 +26,7 @@ else:
     chh = ord
 
 
-class Hex(Blob):
+class Hex(Backend):
     """
     A loader for Intel Hex Objects
     See https://en.wikipedia.org/wiki/Intel_HEX
@@ -69,14 +68,15 @@ class Hex(Blob):
 
         return result
 
-    def _load(self, file_offset, mem_addr, size):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
 
-        # Ignore the params, they don't really work in this format.
         # Do the whole thing in one shot.
+        self.os = 'unknown'
         got_base = False
         got_entry = False
-        self.binary_stream.seek(0)
-        string = self.binary_stream.read()
+        self._binary_stream.seek(0)
+        string = self._binary_stream.read()
         recs = string.splitlines()
         regions = []
         max_addr = 0
@@ -102,7 +102,7 @@ class Hex(Blob):
                 # "Extended Mode" Segment address, take this value, multiply by 16, make the base
                 self._base_address = struct.unpack('>H', data)[0] * 16
                 got_base = True
-                l.debug("Loading a segment at " + hex(self._base_address))
+                l.debug("Loading a segment at %#x", self._base_address)
             elif rectype == HEX_TYPE_STARTSEGADDR:
                 # Four bytes, the segment and the initial IP
                 got_base = True
@@ -110,20 +110,18 @@ class Hex(Blob):
                 self._initial_cs, self._initial_ip = struct.unpack('>HH', data)
                 # The whole thing is the entry, as far as angr is concerned.
                 self._entry = struct.unpack('>I', data)[0]
-                self._custom_entry_point = self._entry
-                l.debug("Got entry point at " + hex(self._entry))
+                l.debug("Got entry point at %#x", self._entry)
             elif rectype == HEX_TYPE_EXTLINEARADDR:
                 got_base = True
                 # Specifies the base for all future data bytes.
                 self._base_address = struct.unpack('>H', data)[0] << 16
-                l.debug("Loading a segment at " + hex(self._base_address))
+                l.debug("Loading a segment at %#x", self._base_address)
             elif rectype == HEX_TYPE_STARTLINEARADDR:
                 got_entry = True
                 # The 32-bit EIP, really the same as STARTSEGADDR, but some compilers pick one over the other.
                 self._entry = struct.unpack('>I', data)[0]
-                l.debug("Found entry point at " + hex(self._entry))
+                l.debug("Found entry point at %#x", self._entry)
                 self._initial_eip = self._entry
-                self._custom_entry_point = self._entry
             else:
                 raise CLEError("This HEX Object type is not implemented: " + hex(rectype))
         if not got_base:
