@@ -13,11 +13,12 @@ class DllImport(PEReloc):
     pass
 
 class IMAGE_REL_BASED_ABSOLUTE(PEReloc):
-    pass
+    def relocate(self):
+        pass
 
 class IMAGE_REL_BASED_HIGHADJ(PEReloc):
     def __init__(self, owner, addr, next_rva):
-        super(IMAGE_REL_BASED_HIGHADJ, self).__init__(owner, None, addr)
+        super().__init__(owner, None, addr)
         self.next_rva = next_rva
     @property
     def value(self):
