@@ -8,7 +8,7 @@ import archinfo
 from archinfo.arch_soot import ArchSoot
 
 from .address_translator import AT
-from .utils import ALIGN_UP, key_bisect_insort_left, key_bisect_floor_key
+from .utils import ALIGN_UP, key_bisect_floor_key, key_bisect_insort_right
 
 try:
     import claripy
@@ -31,6 +31,7 @@ class Loader:
     The following parameters are optional.
 
     :param auto_load_libs:      Whether to automatically load shared libraries that loaded objects depend on.
+    :param load_debug_info:     Whether to automatically parse DWARF data and search for debug symbol files.
     :param concrete_target:     Whether to instantiate a concrete target for a concrete execution of the process.
                                 if this is the case we will need to instantiate a SimConcreteEngine that wraps the
                                 ConcreteTarget provided by the user.
@@ -52,10 +53,6 @@ class Loader:
     :param page_size:           The granularity with which data is mapped into memory. Set to 1 if you are working
                                 in a non-paged environment.
     :param preload_libs:        Similar to `force_load_libs` but will provide for symbol resolution, with precedence
-    :param extern_size:         How much space to allocate for externs. Default 0x8000, or 0x80000 for 64-bit
-                                architectures.
-    :param tls_size:            How much space to allocate for tls. Default 0x8000, or 0x80000 for 64-bit
-                                architectures.
                                 over any dependencies.
     :ivar memory:               The loaded, rebased, and relocated memory of the program.
     :vartype memory:            cle.memory.Clemory
@@ -79,9 +76,9 @@ class Loader:
     def __init__(self, main_binary, auto_load_libs=True, concrete_target = None,
                  force_load_libs=(), skip_libs=(),
                  main_opts=None, lib_opts=None, ld_path=(), use_system_libs=True,
-                 ignore_import_version_numbers=True, case_insensitive=False, rebase_granularity=0x1000000,
-                 except_missing_libs=False, aslr=False, perform_relocations=True,
-                 page_size=0x1, extern_size=None, tls_size=None, preload_libs=(), arch=None):
+                 ignore_import_version_numbers=True, case_insensitive=False, rebase_granularity=0x100000,
+                 except_missing_libs=False, aslr=False, perform_relocations=True, load_debug_info=False,
+                 page_size=0x1, preload_libs=(), arch=None):
         if hasattr(main_binary, 'seek') and hasattr(main_binary, 'read'):
             self._main_binary_path = None
             self._main_binary_stream = main_binary
@@ -89,11 +86,15 @@ class Loader:
             self._main_binary_path = os.path.realpath(str(main_binary))
             self._main_binary_stream = None
 
+        # whether we are presently in the middle of a load cycle
+        self._juggling = False
+
         # auto_load_libs doesn't make any sense if we have a concrete target.
         if concrete_target:
             auto_load_libs = False
 
         self._auto_load_libs = auto_load_libs
+        self._load_debug_info = load_debug_info
         self._satisfied_deps = dict((x, False) for x in skip_libs)
         self._main_opts = {} if main_opts is None else main_opts
         self._lib_opts = {} if lib_opts is None else lib_opts
@@ -105,8 +106,6 @@ class Loader:
         self._case_insensitive = case_insensitive
         self._rebase_granularity = rebase_granularity
         self._except_missing_libs = except_missing_libs
-        self._extern_size = extern_size
-        self._tls_size = tls_size
         self._relocated_objects = set()
         self._perform_relocations = perform_relocations
 
@@ -122,7 +121,7 @@ class Loader:
         self.page_size = page_size
         self.memory = None # type: Clemory
         self.main_object = None # type: Backend
-        self._tls_object = None # type: TLSObject
+        self.tls = None # type: ThreadManager
         self._kernel_object = None # type: KernelObject
         self._extern_object = None # type: ExternObject
         self.shared_objects = OrderedDict()
@@ -131,8 +130,7 @@ class Loader:
         if arch is not None:
             self._main_opts.update({'arch': arch})
         self.preload_libs = []
-        self.initial_load_objects = self._internal_load(main_binary, *preload_libs, preloading=True)
-        self.initial_load_objects.extend(self._internal_load(*force_load_libs))
+        self.initial_load_objects = self._internal_load(main_binary, *preload_libs, *force_load_libs, preloading=(main_binary, *preload_libs))
 
         # cache
         self._last_object = None
@@ -143,11 +141,7 @@ class Loader:
     # Basic functions and properties
 
     def close(self):
-        """
-        Release any resources held by this loader.
-        """
-        for obj in self.all_objects:
-            obj.close()
+        l.warning("You don't need to close the loader anymore :)")
 
     def __repr__(self):
         if self._main_binary_stream is None:
@@ -203,17 +197,25 @@ class Loader:
         Return the extern object used to provide addresses to unresolved symbols and angr internals.
 
         Accessing this property will load this object into memory if it was not previously present.
+
+        proposed model for how multiple extern objects should work:
+
+            1) extern objects are a linked list. the one in loader._extern_object is the head of the list
+            2) each round of explicit loads generates a new extern object if it has unresolved dependencies. this object
+               has exactly the size necessary to hold all its exports.
+            3) All requests for size are passed down the chain until they reach an object which has the space to service it
+               or an object which has not yet been mapped. If all objects have been mapped and are full, a new extern object
+               is mapped with a fixed size.
         """
         if self._extern_object is None:
-            if self._extern_size is None:
-                if self.main_object.arch.bits < 32:
-                    self._extern_size = 0x200
-                elif self.main_object.arch.bits == 32:
-                    self._extern_size = 0x8000
-                else:
-                    self._extern_size = 0x80000
-            self._extern_object = ExternObject(self, map_size=self._extern_size)
-            self._map_object(self._extern_object)
+            if self.main_object.arch.bits < 32:
+                extern_size = 0x200
+            elif self.main_object.arch.bits == 32:
+                extern_size = 0x8000
+            else:
+                extern_size = 0x80000
+            self._extern_object = ExternObject(self, map_size=extern_size)
+            self._internal_load(self._extern_object)
         return self._extern_object
 
     @property
@@ -228,29 +230,6 @@ class Loader:
             self._map_object(self._kernel_object)
         return self._kernel_object
 
-    @property
-    def tls_object(self):
-        """
-        Return the object used to provide addresses for thread-local storage.
-
-        Accessing this property will load this object into memory if it was not previously present.
-        """
-        if self._tls_object is None:
-            if self._tls_size is None:
-                if self.main_object.arch.bits < 32:
-                    self._tls_size = 0x200
-                elif self.main_object.arch.bits == 32:
-                    self._tls_size = 0x8000
-                else:
-                    self._tls_size = 0x80000
-            if isinstance(self.main_object, MetaELF):
-                self._tls_object = ELFTLSObject(self, max_data=self._tls_size)
-                self._map_object(self._tls_object)
-            elif isinstance(self.main_object, PE):
-                self._tls_object = PETLSObject(self, max_data=self._tls_size)
-                self._map_object(self._tls_object)
-        return self._tls_object
-
     @property
     def all_elf_objects(self):
         """
@@ -326,6 +305,12 @@ class Loader:
         """
         If the given library specification has been loaded, return its object, otherwise return None.
         """
+        if isinstance(spec, Backend):
+            for obj in self.all_objects:
+                if obj is spec:
+                    return obj
+            return None
+
         if self._case_insensitive:
             spec = spec.lower()
         extra_idents = {}
@@ -368,6 +353,7 @@ class Loader:
         if self._last_object is not None and \
                 self._last_object.min_addr <= addr <= self._last_object.max_addr:
             if not membership_check: return self._last_object
+            if not self._last_object.has_memory: return self._last_object
             o = _check_object_memory(self._last_object)
             if o: return o
 
@@ -380,8 +366,10 @@ class Loader:
         if not obj.min_addr <= addr <= obj.max_addr:
             return None
         if not membership_check:
+            self._last_object = obj
             return obj
         if not obj.has_memory:
+            self._last_object = obj
             return obj
         return _check_object_memory(obj)
 
@@ -601,16 +589,6 @@ class Loader:
             l.warning("Dynamic load failed: %r", e)
             return None
 
-    def add_object(self, obj):
-        """
-        If you've constructed your own Backend-subclass object and want to add it directly to the loader, use this.
-        """
-        self._register_object(obj)
-        self._map_object(obj)
-        if isinstance(obj, (MetaELF, PE)) and obj.tls_used:
-            self.tls_object.register_object(obj)
-        self._relocate_object(obj)
-
     def get_loader_symbolic_constraints(self):
         """
         Do not use this method.
@@ -638,7 +616,7 @@ class Loader:
         """
         return 'ld.so' in name or 'ld64.so' in name or 'ld-linux' in name
 
-    def _internal_load(self, *args, preloading=False):
+    def _internal_load(self, *args, preloading=()):
         """
         Pass this any number of files or libraries to load. If it can't load any of them for any reason, it will
         except out. Note that the semantics of ``auto_load_libs`` and ``except_missing_libs`` apply at all times.
@@ -646,86 +624,168 @@ class Loader:
         It will return a list of all the objects successfully loaded, which may be smaller than the list you provided
         if any of them were previously loaded.
 
-        The ``main_binary`` has to come first, followed by any ``preload_libs`` to ensure symbols are resolved to
-        preloaded libraries ahead of any others. Without ``preloading=True``, then all the libraries will be treated
-        like ``force_load_libs`` (i.e., not resolving symbols to them).
-
-        :param preloading: If True, loaded objects are appended to ``self.preload_libs`` (i.e., force_load_libs are
-                           not treated like preload_libs)
+        The ``main_binary`` has to come first, followed by any additional libraries to load this round. To create the
+        effect of "preloading", i.e. ensuring symbols are resolved to preloaded libraries ahead of any others, pass
+        ``preloading`` as a list of identifiers which should be considered preloaded. Note that the identifiers will
+        be compared using object identity.
         """
+        # ideal loading pipeline:
+        # - load everything, independently and recursively until dependencies are satisfied
+        # - resolve symbol-based dependencies
+        # - layout address space, including (as a prerequisite) coming up with the layout for tls and externs
+        # - map everything into memory
+        # - perform relocations
+
+        # STEP 1
+        # Load everything. for each binary, load it in isolation so we end up with a Backend instance.
+        # If auto_load_libs is on, do this iteratively until all dependencies is satisfied
         objects = []
+        preload_objects = []
         dependencies = []
         cached_failures = set() # this assumes that the load path is global and immutable by the time we enter this func
 
         for main_spec in args:
+            is_preloading = any(spec is main_spec for spec in preloading)
             if self.find_object(main_spec, extra_objects=objects) is not None:
                 l.info("Skipping load request %s - already loaded", main_spec)
                 continue
-            main_obj = self._load_object_isolated(main_spec)
-            objects.append(main_obj)
-            dependencies.extend(main_obj.deps)
+            obj = self._load_object_isolated(main_spec)
+            objects.append(obj)
+            objects.extend(obj.child_objects)
+            dependencies.extend(obj.deps)
 
             if self.main_object is None:
-                self.main_object = main_obj
-                self.memory = Clemory(self.main_object.arch, root=True)
-            elif preloading:
-                    self.preload_libs.append(main_obj)
+                # this is technically the first place we can start to initialize things based on platform
+                self.main_object = obj
+                self.memory = Clemory(obj.arch, root=True)
+
+                chk_obj = self.main_object if isinstance(self.main_object, ELFCore) or not self.main_object.child_objects else self.main_object.child_objects[0]
+                if isinstance(chk_obj, ELFCore):
+                    self.tls = ELFCoreThreadManager(self, obj.arch)
+                elif isinstance(obj, Minidump):
+                    self.tls = MinidumpThreadManager(self, obj.arch)
+                elif isinstance(chk_obj, MetaELF):
+                    self.tls = ELFThreadManager(self, obj.arch)
+                elif isinstance(chk_obj, PE):
+                    self.tls = PEThreadManager(self, obj.arch)
+                else:
+                    self.tls = ThreadManager(self, obj.arch)
+
+            elif is_preloading:
+                self.preload_libs.append(obj)
+                preload_objects.append(obj)
+
 
         while self._auto_load_libs and dependencies:
-            dep_spec = dependencies.pop(0)
-            if dep_spec in cached_failures:
-                l.debug("Skipping implicit dependency %s - cached failure", dep_spec)
+            spec = dependencies.pop(0)
+            if spec in cached_failures:
+                l.debug("Skipping implicit dependency %s - cached failure", spec)
                 continue
-            if self.find_object(dep_spec, extra_objects=objects) is not None:
-                l.debug("Skipping implicit dependency %s - already loaded", dep_spec)
+            if self.find_object(spec, extra_objects=objects) is not None:
+                l.debug("Skipping implicit dependency %s - already loaded", spec)
                 continue
 
             try:
-                l.info("Loading %s...", dep_spec)
-                dep_obj = self._load_object_isolated(dep_spec)  # loading dependencies
+                l.info("Loading %s...", spec)
+                obj = self._load_object_isolated(spec)  # loading dependencies
             except CLEFileNotFoundError:
                 l.info("... not found")
-                cached_failures.add(dep_spec)
+                cached_failures.add(spec)
                 if self._except_missing_libs:
                     raise
-                else:
-                    continue
+                continue
 
-            objects.append(dep_obj)
-            dependencies.extend(dep_obj.deps)
+            objects.append(obj)
+            objects.extend(obj.child_objects)
+            dependencies.extend(obj.deps)
 
+            if type(self.tls) is ThreadManager:   # ... java
+                if isinstance(obj, MetaELF):
+                    self.tls = ELFThreadManager(self, obj.arch)
+                elif isinstance(obj, PE):
+                    self.tls = PEThreadManager(self, obj.arch)
+
+        # STEP 1.5
+        # produce dependency-ordered list of objects and soname map
+
+        ordered_objects = []
+        soname_mapping = OrderedDict((obj.provides, obj) for obj in objects if obj.provides)
+        seen = set()
+        def visit(obj):
+            if id(obj) in seen:
+                return
+            seen.add(id(obj))
+
+            dep_objs = [soname_mapping[dep_name] for dep_name in obj.deps if dep_name in soname_mapping]
+            for dep_obj in dep_objs:
+                visit(dep_obj)
+
+            ordered_objects.append(obj)
+
+        for obj in preload_objects + objects:
+            visit(obj)
+
+        # STEP 2
+        # Resolve symbol dependencies. Create an unmapped extern object, which may not be used
+        # after this step, everything should have the appropriate references to each other and the extern
+        # object should have all the space it needs allocated
+
+        extern_obj = ExternObject(self)
+
+        # tls registration
         for obj in objects:
-            self._register_object(obj)
+            self.tls.register_object(obj)
+
+        # link everything
+        if self._perform_relocations:
+            for obj in ordered_objects:
+                l.info("Linking %s", obj.binary)
+                sibling_objs = list(obj.parent_object.child_objects) if obj.parent_object is not None else []
+                dep_objs = [soname_mapping[dep_name] for dep_name in obj.deps if dep_name in soname_mapping]
+                main_objs = [self.main_object] if self.main_object is not obj else []
+                for reloc in obj.relocs:
+                    reloc.resolve_symbol(main_objs + preload_objects + sibling_objs + dep_objs + [obj], extern_object=extern_obj)
+
+        # if the extern object was used, add it to the list of objects we're mapping
+        # also add it to the linked list of extern objects
+        if extern_obj.map_size:
+            # resolve the extern relocs this way because they may produce more relocations as we go
+            i = 0
+            while i < len(extern_obj.relocs):
+                extern_obj.relocs[i].resolve_symbol(objects, extern_object=extern_obj)
+                i += 1
+
+            objects.append(extern_obj)
+            ordered_objects.insert(0, extern_obj)
+            extern_obj._next_object = self._extern_object
+            self._extern_object = extern_obj
+
+            extern_obj._finalize_tls()
+            self.tls.register_object(extern_obj)
+
+        # STEP 3
+        # Map everything to memory
         for obj in objects:
             self._map_object(obj)
-        for obj in objects:
-            if isinstance(obj, (MetaELF, PE)) and obj.tls_used:
-                self.tls_object.register_object(obj)
-        if self._perform_relocations:
-            for obj in objects:
-                self._relocate_object(obj)
 
+        # STEP 4
+        # Perform relocations
+        if self._perform_relocations:
+            for obj in ordered_objects:
+                obj.relocate()
+
+        # Step 5
+        # Insert each object into the appropriate mappings for lookup by name
         for obj in objects:
-            if isinstance(obj, (MetaELF, PE)) and obj.tls_used:
-                self.tls_object.map_object(obj)
-        if self._extern_object and self._extern_object.tls_used and not self._extern_object._tls_mapped:
-            # this entire scheme will break when we do dynamic loading. you have been warned, me.
-            self.tls_object.map_object(self._extern_object)
-            self._extern_object._tls_mapped = True
+            self.requested_names.update(obj.deps)
+            for ident in self._possible_idents(obj):
+                self._satisfied_deps[ident] = obj
+
+            if obj.provides is not None:
+                self.shared_objects[obj.provides] = obj
 
         return objects
 
-    def _register_object(self, obj):
-        """
-        Insert this object's clerical information into the loader
-        """
-        self.requested_names.update(obj.deps)
-        for ident in self._possible_idents(obj):
-            self._satisfied_deps[ident] = obj
-
-        if obj.provides is not None:
-            self.shared_objects[obj.provides] = obj
-
     def _load_object_isolated(self, spec):
         """
         Given a partial specification of a dependency, this will return the loaded object as a backend instance.
@@ -735,36 +795,46 @@ class Loader:
         if isinstance(spec, Backend):
             return spec
         elif hasattr(spec, 'read') and hasattr(spec, 'seek'):
-            full_spec = spec
+            binary_stream = spec
+            binary = None
+            close = False
         elif type(spec) in (bytes, str):
-            full_spec = self._search_load_path(spec) # this is allowed to cheat and do partial static loading
-            l.debug("... using full path %s", full_spec)
+            binary = self._search_load_path(spec) # this is allowed to cheat and do partial static loading
+            l.debug("... using full path %s", binary)
+            binary_stream = open(binary, 'rb')
+            close = True
         else:
             raise CLEError("Bad library specification: %s" % spec)
 
-        # STEP 2: collect options
-        if self.main_object is None:
-            options = self._main_opts
-        else:
-            for ident in self._possible_idents(full_spec): # also allowed to cheat
-                if ident in self._lib_opts:
-                    options = self._lib_opts[ident]
-                    break
+        try:
+            # STEP 2: collect options
+            if self.main_object is None:
+                options = dict(self._main_opts)
             else:
-                options = {}
+                for ident in self._possible_idents(binary_stream if binary is None else binary): # also allowed to cheat
+                    if ident in self._lib_opts:
+                        options = dict(self._lib_opts[ident])
+                        break
+                else:
+                    options = {}
 
-        # STEP 3: identify backend
-        backend_spec = options.pop('backend', None)
-        backend_cls = self._backend_resolver(backend_spec)
-        if backend_cls is None:
-            backend_cls = self._static_backend(full_spec)
-        if backend_cls is None:
-            raise CLECompatibilityError("Unable to find a loader backend for %s.  Perhaps try the 'blob' loader?" % spec)
+            # STEP 3: identify backend
+            backend_spec = options.pop('backend', None)
+            backend_cls = self._backend_resolver(backend_spec)
+            if backend_cls is None:
+                backend_cls = self._static_backend(binary_stream if binary is None else binary)
+            if backend_cls is None:
+                raise CLECompatibilityError("Unable to find a loader backend for %s.  Perhaps try the 'blob' loader?" % spec)
 
-        # STEP 4: LOAD!
-        l.debug("... loading with %s", backend_cls)
+            # STEP 4: LOAD!
+            l.debug("... loading with %s", backend_cls)
 
-        return backend_cls(full_spec, is_main_bin=self.main_object is None, loader=self, **options)
+            result = backend_cls(binary, binary_stream, is_main_bin=self.main_object is None, loader=self, **options)
+            result.close()
+            return result
+        finally:
+            if close:
+                binary_stream.close()
 
     def _map_object(self, obj):
         """
@@ -784,45 +854,23 @@ class Loader:
                           "It is being loaded with a base address of 0x400000.")
                 base_addr = 0x400000
 
-            obj.mapped_base = base_addr
-            obj.rebase()
+            obj.rebase(base_addr)
         else:
             if obj._custom_base_addr is not None and not isinstance(obj, Blob):
                 l.warning("%s: base_addr was specified but the object is not PIC. "
-                    "specify force_rebase=True to override",
-                            os.path.basename(obj.binary) if obj.binary is not None else obj.binary_stream)
+                          "specify force_rebase=True to override", obj.binary_basename)
             base_addr = obj.linked_base
             if not self._is_range_free(obj.linked_base, obj_size):
                 raise CLEError("Position-DEPENDENT object %s cannot be loaded at %#x"% (obj.binary, base_addr))
 
-        assert obj.min_addr < obj.max_addr
         assert obj.mapped_base >= 0
 
         if obj.has_memory:
+            assert obj.min_addr < obj.max_addr
             l.info("Mapping %s at %#x", obj.binary, base_addr)
             self.memory.add_backer(base_addr, obj.memory)
-        key_bisect_insort_left(self.all_objects, obj, keyfunc=lambda o: o.min_addr)
         obj._is_mapped = True
-
-    def _relocate_object(self, obj):
-        """
-        Perform the relocations for ``obj``, making sure its dependencies are relocated first
-        """
-        if id(obj) in self._relocated_objects:
-            return
-        self._relocated_objects.add(id(obj))
-
-        for f in self.preload_libs:
-            self._relocate_object(f)
-
-        dep_objs = [self.shared_objects[dep_name] for dep_name in obj.deps if dep_name in self.shared_objects]
-        for dep_obj in dep_objs:
-            self._relocate_object(dep_obj)
-
-        l.info("Relocating %s", obj.binary)
-        for reloc in obj.relocs:
-            if not reloc.resolved:
-                reloc.relocate(([self.main_object] if self.main_object is not obj else []) + self.preload_libs + dep_objs + [obj])
+        key_bisect_insort_right(self.all_objects, obj, keyfunc=lambda o: o.min_addr)
 
     # Address space management
 
@@ -1074,7 +1122,7 @@ class Loader:
 
 from .errors import CLEError, CLEFileNotFoundError, CLECompatibilityError, CLEOperationError
 from .memory import Clemory
-from .backends import MetaELF, ELF, PE, Blob, ALL_BACKENDS, Backend
-from .backends.tls import PETLSObject, ELFTLSObject, TLSObject
+from .backends import MetaELF, ELF, PE, ELFCore, Minidump, Blob, ALL_BACKENDS, Backend
+from .backends.tls import ThreadManager, ELFThreadManager, PEThreadManager, ELFCoreThreadManager, MinidumpThreadManager, TLSObject
 from .backends.externs import ExternObject, KernelObject
 from .utils import stream_or_path
